"""Tasmota utility functions."""
from __future__ import annotations

import json
import logging
import operator
import re
from collections.abc import Mapping
from functools import reduce
from typing import Any, Dict, cast

from .const import (
    CONF_DEVICENAME,
    CONF_FRIENDLYNAME,
    CONF_FULLTOPIC,
    CONF_HOSTNAME,
    CONF_MAC,
    CONF_OFFLINE,
    CONF_ONLINE,
    CONF_PREFIX,
    CONF_STATE,
    CONF_SWITCHNAME,
    CONF_TOPIC,
    PREFIX_CMND,
    PREFIX_STAT,
    PREFIX_TELE,
    RSLT_ACTION,
    RSLT_POWER,
    RSLT_STATE,
    STATE_OFF,
    STATE_ON,
)
from .mqtt import ReceivePayloadType

_LOGGER = logging.getLogger(__name__)

ConfigType = Dict[str, str]


def get_by_path(root: dict, items: list[str | int]) -> dict:
    """Access a nested object in root by item sequence."""
    return reduce(operator.getitem, items, root)


def set_by_path(root: dict, items: list[str | int], value: Any) -> None:
    """Set a value in a nested object in root by item sequence."""
    get_by_path(root, items[:-1])[items[-1]] = value


def del_by_path(root: dict, items: list[str | int]) -> None:
    """Delete a key-value in a nested object in root by item sequence."""
    del get_by_path(root, items[:-1])[items[-1]]


def _get_topic(config: ConfigType, prefix: str) -> str:
    topic = config[CONF_FULLTOPIC]
    topic = topic.replace("%hostname%", config[CONF_HOSTNAME])
    topic = topic.replace("%id%", config[CONF_MAC][-6:])
    topic = topic.replace("%prefix%", prefix)
    topic = topic.replace("%topic%", config[CONF_TOPIC])
    return topic


def _get_topic_cmnd(config: ConfigType) -> str:
    return _get_topic(config, config[CONF_PREFIX][PREFIX_CMND])


def _get_topic_stat(config: ConfigType) -> str:
    return _get_topic(config, config[CONF_PREFIX][PREFIX_STAT])


def _get_topic_tele(config: ConfigType) -> str:
    return _get_topic(config, config[CONF_PREFIX][PREFIX_TELE])


def get_topic_command(config: ConfigType) -> str:
    """Get command topic."""
    return _get_topic_cmnd(config)


def get_topic_command_state(config: ConfigType) -> str:
    """Get topic for command power."""
    return _get_topic_cmnd(config) + "STATE"


def get_topic_command_status(config: ConfigType) -> str:
    """Get topic for command power."""
    return _get_topic_cmnd(config) + "STATUS"


def get_topic_stat(config: ConfigType) -> str:
    """Get stat topic."""
    return _get_topic_stat(config)


def get_topic_stat_button_trigger(config: ConfigType, idx: int) -> str:
    """Get topic for tele state."""
    return _get_topic_stat(config) + f"BUTTON{idx+1}"


def get_topic_stat_result(config: ConfigType) -> str:
    """Get topic for tele state."""
    return _get_topic_stat(config) + "RESULT"


def get_topic_stat_status(config: ConfigType, idx: int | None = None) -> str:
    """Get topic for tele state."""
    if idx is None:
        return _get_topic_stat(config) + "STATUS"
    return _get_topic_stat(config) + f"STATUS{idx}"


def get_topic_stat_switch(config: ConfigType, idx: int) -> str:
    """Get topic for tele state."""
    return _get_topic_stat(config) + f"SWITCH{idx+1}"


def get_topic_stat_switch_trigger(config: ConfigType, idx: int) -> str:
    """Get topic for tele state."""
    return _get_topic_stat(config) + f"SWITCH{idx+1}"


def get_topic_tele(config: ConfigType) -> str:
    """Get tele topic."""
    return _get_topic_tele(config)


def get_topic_tele_sensor(config: ConfigType) -> str:
    """Get topic for tele state."""
    return _get_topic_tele(config) + "SENSOR"


def get_topic_tele_state(config: ConfigType) -> str:
    """Get topic for tele state."""
    return _get_topic_tele(config) + "STATE"


def get_topic_tele_will(config: ConfigType) -> str:
    """Get topic for tele will."""
    return _get_topic_tele(config) + "LWT"


def config_get_state_power_on(config: ConfigType) -> str:
    """Get command/result on."""
    return config[CONF_STATE][STATE_ON]


def config_get_state_power_off(config: ConfigType) -> str:
    """Get command/result off."""
    return config[CONF_STATE][STATE_OFF]


def config_get_state_offline(config: ConfigType) -> str:
    """Get state offline."""
    return config[CONF_OFFLINE]


def config_get_state_online(config: ConfigType) -> str:
    """Get state online."""
    return config[CONF_ONLINE]


def get_value(
    _status: str, key: str, idx: int | None = None, idx_optional: bool = False
) -> Any:
    """Get status from JSON formatted status or result."""
    try:
        status = json.loads(_status)
    except json.decoder.JSONDecodeError:
        _LOGGER.debug("Invalid JSON '%s'", _status)
        return None
    if idx is None:
        return status.get(key)
    if key in status and idx_optional and idx == 0:
        return status[key]
    key = f"{key}{idx+1}"
    return status[key] if key in status else None


def get_value_by_path(status: dict | ReceivePayloadType, path: list[str | int]) -> Any:
    """Get status from JSON formatted status or result by path."""
    try:
        if not isinstance(status, Mapping):
            status = json.loads(status)
        return get_by_path(cast(dict, status), path)
    except (json.decoder.JSONDecodeError, KeyError):
        return None


def get_state_power(status: str, idx: int) -> Any:
    """Get state power."""
    return get_value(status, RSLT_POWER, idx=idx, idx_optional=True)


def get_state_state(status: str) -> Any:
    """Get state of switch."""
    return get_value(status, RSLT_STATE)


def get_state_button_trigger(status: str) -> Any:
    """Get state of button."""
    return get_value(status, RSLT_ACTION)


def config_get_friendlyname(config: ConfigType, platform: str, idx: int) -> str:
    """Get config friendly name."""
    friendly_names = config[CONF_FRIENDLYNAME]

    if idx >= len(friendly_names) or friendly_names[idx] is None:
        return f"{config[CONF_DEVICENAME]} {platform} {idx+1}"
    return friendly_names[idx]


def config_get_switchfriendlyname(config: ConfigType, platform: str, idx: int) -> str:
    """Get config friendly name."""
    switch_names = config[CONF_SWITCHNAME]

    if idx >= len(switch_names) or switch_names[idx] is None:
        return f"{config[CONF_DEVICENAME]} {platform} {idx+1}"
    return switch_names[idx]


def config_get_switchname(config: ConfigType, idx: int) -> str:
    """Get switch name."""
    switch_names = config[CONF_SWITCHNAME]

    if idx >= len(switch_names) or switch_names[idx] is None:
        return f"Switch{idx+1}"
    return switch_names[idx]


TOPIC_MATCHER = re.compile(r"^(?P<mac>[A-Z0-9_-]+)\/(?:config|sensors)$")


def discovery_topic_get_mac(topic: str, discovery_topic: str) -> str | None:
    """Get MAC from discovery topic."""
    topic_trimmed = topic.replace(f"{discovery_topic}/", "", 1)
    match = TOPIC_MATCHER.match(topic_trimmed)

    if not match:
        return None

    (mac,) = match.groups()
    return mac


def discovery_topic_is_device_config(topic: str) -> bool:
    """Return True if the discovery topic is device configuration."""
    return topic.endswith("config")



"""Tasmota shutter."""
from __future__ import annotations

import logging
from typing import Any

import attr

from .const import (
    COMMAND_SHUTTER_CLOSE,
    COMMAND_SHUTTER_OPEN,
    COMMAND_SHUTTER_POSITION,
    COMMAND_SHUTTER_STOP,
    COMMAND_SHUTTER_TILT,
    CONF_DEVICENAME,
    CONF_MAC,
    CONF_SHUTTER_OPTIONS,
    CONF_SHUTTER_TILT,
    RSLT_SHUTTER,
    SHUTTER_DIRECTION,
    SHUTTER_OPTION_INVERT,
    SHUTTER_POSITION,
    SHUTTER_TILT,
    STATUS_SENSOR,
)
from .entity import (
    TasmotaAvailability,
    TasmotaAvailabilityConfig,
    TasmotaEntity,
    TasmotaEntityConfig,
)
from .mqtt import ReceiveMessage
from .utils import (
    config_get_state_offline,
    config_get_state_online,
    get_topic_command,
    get_topic_command_status,
    get_topic_stat_result,
    get_topic_stat_status,
    get_topic_tele_sensor,
    get_topic_tele_will,
    get_value_by_path,
)

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True, frozen=True)
class TasmotaShutterConfig(TasmotaAvailabilityConfig, TasmotaEntityConfig):
    """Tasmota shutter configuation."""

    idx: int = attr.ib()
    command_topic: str = attr.ib()
    inverted_shutter: bool = attr.ib()
    state_topic1: str = attr.ib()
    state_topic2: str = attr.ib()
    state_topic3: str = attr.ib()
    tilt_min: int = attr.ib()
    tilt_max: int = attr.ib()
    tilt_dur: int = attr.ib()

    @classmethod
    def from_discovery_message(
        cls, config: dict, idx: int, platform: str
    ) -> TasmotaShutterConfig:
        """Instantiate from discovery message."""
        shutter_options = config[CONF_SHUTTER_OPTIONS]
        shutter_options = shutter_options[idx] if idx < len(shutter_options) else 0
        shutter_tilt = config[CONF_SHUTTER_TILT]
        shutter_tilt = shutter_tilt[idx] if idx < len(shutter_tilt) else [0, 0, 0]
        return cls(
            endpoint="shutter",
            idx=idx,
            friendly_name=f"{config[CONF_DEVICENAME]} {platform} {idx+1}",
            mac=config[CONF_MAC],
            platform=platform,
            poll_payload="10",
            poll_topic=get_topic_command_status(config),
            availability_topic=get_topic_tele_will(config),
            availability_offline=config_get_state_offline(config),
            availability_online=config_get_state_online(config),
            command_topic=get_topic_command(config),
            inverted_shutter=shutter_options & SHUTTER_OPTION_INVERT,
            state_topic1=get_topic_stat_result(config),
            state_topic2=get_topic_tele_sensor(config),
            state_topic3=get_topic_stat_status(config, 10),
            tilt_min=shutter_tilt[0],
            tilt_max=shutter_tilt[1],
            tilt_dur=shutter_tilt[2],
        )


class TasmotaShutter(TasmotaAvailability, TasmotaEntity):
    """Representation of a Tasmota shutter."""

    _cfg: TasmotaShutterConfig

    def __init__(self, **kwds: Any):
        """Initialize."""
        self._sub_state: dict | None = None
        super().__init__(**kwds)

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

        def state_message_received(msg: ReceiveMessage) -> None:
            """Handle new MQTT state messages."""
            if not self._on_state_callback:
                return

            shutter = f"{RSLT_SHUTTER}{self._cfg.idx+1}"
            prefix: list[str | int] = []
            if msg.topic == self._cfg.state_topic3:
                prefix = [STATUS_SENSOR]

            direction = get_value_by_path(
                msg.payload, prefix + [shutter, SHUTTER_DIRECTION]
            )
            if direction is not None and self._cfg.inverted_shutter:
                direction = direction * -1

            position = get_value_by_path(
                msg.payload, prefix + [shutter, SHUTTER_POSITION]
            )
            if position is not None and self._cfg.inverted_shutter:
                position = 100 - position

            tilt = get_value_by_path(msg.payload, prefix + [shutter, SHUTTER_TILT])
            ha_tilt = None
            if tilt is not None:
                ha_tilt_range = 100
                tasmota_tilt_range = self._cfg.tilt_max - self._cfg.tilt_min
                if tasmota_tilt_range:
                    ha_tilt = (
                        (tilt - self._cfg.tilt_min) * ha_tilt_range / tasmota_tilt_range
                    )

            if direction is not None or position is not None or ha_tilt is not None:
                self._on_state_callback(
                    None, direction=direction, position=position, tilt=ha_tilt
                )

        availability_topics = self.get_availability_topics()
        topics = {
            "state_topic1": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic1,
                "msg_callback": state_message_received,
            },
            "state_topic2": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic2,
                "msg_callback": state_message_received,
            },
            "state_topic3": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic3,
                "msg_callback": state_message_received,
            },
        }
        topics = {**topics, **availability_topics}

        self._sub_state = await self._mqtt_client.subscribe(
            self._sub_state,
            topics,
        )

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe to all MQTT topics."""
        self._sub_state = await self._mqtt_client.unsubscribe(self._sub_state)

    async def open(self) -> None:
        """Open the shutter."""
        payload = ""
        command = f"{COMMAND_SHUTTER_OPEN}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def close(self) -> None:
        """Close the shutter."""
        payload = ""
        command = f"{COMMAND_SHUTTER_CLOSE}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def set_position(self, position: int) -> None:
        """Set the shutter's position.

        0 is closed, 100 is fully open.
        """
        if self._cfg.inverted_shutter:
            position = 100 - position
        payload = position
        command = f"{COMMAND_SHUTTER_POSITION}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def stop(self) -> None:
        """Stop the shutter."""
        payload = ""
        command = f"{COMMAND_SHUTTER_STOP}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    @property
    def supports_tilt(self) -> bool:
        """Return if the shutter supports tilt."""
        return self._cfg.tilt_dur != 0 and (self._cfg.tilt_min != self._cfg.tilt_max)

    async def open_tilt(self) -> None:
        """Open the shutter tilt."""
        payload = "OPEN"
        command = f"{COMMAND_SHUTTER_TILT}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def close_tilt(self) -> None:
        """Close the shutter tilt."""
        payload = "CLOSE"
        command = f"{COMMAND_SHUTTER_TILT}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def set_tilt_position(self, tilt: int) -> None:
        """Set the shutter's tilt position.

        0 is closed, 100 is fully open.
        """
        ha_tilt_range = 100
        tasmota_tilt_range = self._cfg.tilt_max - self._cfg.tilt_min
        tasmota_tilt = self._cfg.tilt_min + (tilt * tasmota_tilt_range / ha_tilt_range)
        payload = round(tasmota_tilt)
        command = f"{COMMAND_SHUTTER_TILT}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )



"""Tasmota types."""
from typing import Awaitable, Callable, List, Tuple, TypedDict, Union

import attr

from .entity import TasmotaAvailabilityConfig, TasmotaEntityConfig


@attr.s(slots=True, frozen=True)
class TasmotaBaseSensorConfig(TasmotaAvailabilityConfig, TasmotaEntityConfig):
    """Tasmota Base Sensor configuration."""


DiscoveryHashType = Tuple[str, str, str, Union[str, int]]
DeviceDiscoveredCallback = Callable[[dict, str], Awaitable[None]]
SensorsDiscoveredCallback = Callable[
    [List[Tuple[TasmotaBaseSensorConfig, DiscoveryHashType]], str], Awaitable[None]
]


class TasmotaDeviceConfig(TypedDict, total=False):
    """Tasmota device config."""

    ip: str
    mac: str
    manufacturer: str
    md: str
    name: str
    sw: str



"""Tasmota discovery."""
from __future__ import annotations

import logging
from typing import Any, Awaitable, Callable

import attr

from .mqtt import ReceiveMessage, TasmotaMQTTClient

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True, frozen=True)
class TasmotaEntityConfig:
    """Base class for Tasmota configuation."""

    endpoint: str = attr.ib()
    idx: int | str | None = attr.ib()
    friendly_name: str | None = attr.ib()
    mac: str = attr.ib()
    platform: str = attr.ib()
    poll_payload: str = attr.ib()
    poll_topic: str = attr.ib()

    @property
    def unique_id(self) -> str:
        """Return unique_id."""
        return f"{self.mac}_{self.platform}_{self.endpoint}_{self.idx}"


@attr.s(slots=True, frozen=True)
class TasmotaAvailabilityConfig(TasmotaEntityConfig):
    """Tasmota availability configuation."""

    availability_topic: str = attr.ib()
    availability_offline: str = attr.ib()
    availability_online: str = attr.ib()


class TasmotaEntity:
    """Base class for Tasmota entities."""

    def __init__(self, config: TasmotaEntityConfig, mqtt_client: TasmotaMQTTClient):
        """Initialize."""
        self._cfg = config
        self._mqtt_client = mqtt_client
        self._on_state_callback: Callable | None = None
        super().__init__()

    def config_same(self, new_config: TasmotaEntityConfig) -> bool:
        """Return if updated config is same as current config."""
        return self._cfg == new_config

    def config_update(self, new_config: TasmotaEntityConfig) -> None:
        """Update config."""
        self._cfg = new_config

    async def poll_status(self) -> None:
        """Poll for status."""
        await self._mqtt_client.publish_debounced(
            self._cfg.poll_topic, self._cfg.poll_payload
        )

    def set_on_state_callback(self, on_state_callback: Callable) -> None:
        """Set callback for state change."""
        self._on_state_callback = on_state_callback

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe to all MQTT topics."""

    @property
    def mac(self) -> str:
        """Return MAC."""
        return self._cfg.mac

    @property
    def name(self) -> str | None:
        """Return friendly name."""
        return self._cfg.friendly_name

    @property
    def unique_id(self) -> str:
        """Return unique_id."""
        return self._cfg.unique_id


class TasmotaAvailability(TasmotaEntity):
    """Availability mixin for Tasmota entities."""

    _cfg: TasmotaAvailabilityConfig

    def __init__(self, **kwds: Any):
        """Initialize."""
        self._on_availability_callback: Awaitable | None = None
        super().__init__(**kwds)

    def get_availability_topics(self) -> dict:
        """Return MQTT topics to subscribe to for availability state."""

        async def availability_message_received(msg: ReceiveMessage) -> None:
            """Handle a new received MQTT availability message."""
            if msg.payload == self._cfg.availability_online:
                await self.poll_status()
            if not self._on_availability_callback:
                return
            if msg.payload == self._cfg.availability_online:
                await self._on_availability_callback(True)
            if msg.payload == self._cfg.availability_offline:
                await self._on_availability_callback(False)

        topics = {
            "availability_topic": {
                "event_loop_safe": True,
                "msg_callback": availability_message_received,
                "topic": self._cfg.availability_topic,
            }
        }
        return topics

    def set_on_availability_callback(self, on_availability_callback: Callable) -> None:
        """Set callback for availability state change."""
        self._on_availability_callback = on_availability_callback



"""Tasmota binary sensor."""
from __future__ import annotations

import logging

import attr

from .const import (
    CONF_BUTTON,
    CONF_MAC,
    CONF_OPTIONS,
    OPTION_BUTTON_SINGLE,
    OPTION_BUTTON_SWAP,
    OPTION_MQTT_BUTTONS,
    RSLT_ACTION,
)
from .mqtt import ReceiveMessage
from .trigger import TasmotaTrigger, TasmotaTriggerConfig
from .utils import get_topic_stat_result, get_value_by_path

_LOGGER = logging.getLogger(__name__)

# Button matrix for triggers generation when SetOption73 is enabled:
# N  SetOption1  SetOption11 SetOption13 SINGLE PRESS                        DOUBLE PRESS  MULTI PRESS                     HOLD
# 1  0           0           0           SINGLE (10 - button_short_press)    DOUBLE        DOUBLE to PENTA                 YES (button_long_press)
# 2  1           0           0           SINGLE (10 - button_short_press)    DOUBLE        DOUBLE to PENTA                 YES (button_long_press)
# 3  0           1           0           DOUBLE (11 - button_short_press)    SINGLE        SINGLE then TRIPLE TO PENTA     YES (button_long_press)
# 4  1           1           0           DOUBLE (11 - button_short_press)    SINGLE        SINGLE then TRIPLE TO PENTA     YES (button_long_press)
# 5  0           0           1           SINGLE (10 - button_short_press)    NONE          NONE                            NONE
# 6  1           0           1           SINGLE (10 - button_short_press)    NONE          NONE                            NONE
# 7  0           1           1           SINGLE (10 - button_short_press)    NONE          NONE                            NONE
# 8  1           1           1           SINGLE (10 - button_short_press)    NONE          NONE                            NONE
# Trigger types:  10 = button_short_press | 11 = button_double_press | 12 = button_triple_press | 13 = button_quadruple_press | 14 = button_quintuple_press | 3 = button_long_press

# SetOption11: Swap button single and double press functionality
# SetOption13: Immediate action on button press, just SINGLE trigger

BUTTONMODE_NONE = "none"
BUTTONMODE_NORMAL = "normal"
BUTTONMODE_SWAP = "swap"
BUTTONMODE_SINGLE = "single"

BTN_SINGLE = "SINGLE"
BTN_DOUBLE = "DOUBLE"
BTN_TRIPLE = "TRIPLE"
BTN_QUAD = "QUAD"
BTN_PENTA = "PENTA"
BTN_HOLD = "HOLD"

BTN_TRIG_NONE = "none"
BTN_TRIG_SINGLE = "button_short_press"
BTN_TRIG_DOUBLE = "button_double_press"
BTN_TRIG_TRIPLE = "button_triple_press"
BTN_TRIG_QUAD = "button_quadruple_press"
BTN_TRIG_PENTA = "button_quintuple_press"
BTN_TRIG_HOLD = "button_long_press"

BUTTONMODE_MAP = {
    BUTTONMODE_NONE: {
        BTN_SINGLE: BTN_TRIG_NONE,
        BTN_DOUBLE: BTN_TRIG_NONE,
        BTN_TRIPLE: BTN_TRIG_NONE,
        BTN_QUAD: BTN_TRIG_NONE,
        BTN_PENTA: BTN_TRIG_NONE,
        BTN_HOLD: BTN_TRIG_NONE,
    },
    BUTTONMODE_NORMAL: {
        BTN_SINGLE: BTN_TRIG_SINGLE,
        BTN_DOUBLE: BTN_TRIG_DOUBLE,
        BTN_TRIPLE: BTN_TRIG_TRIPLE,
        BTN_QUAD: BTN_TRIG_QUAD,
        BTN_PENTA: BTN_TRIG_PENTA,
        BTN_HOLD: BTN_TRIG_HOLD,
    },
    BUTTONMODE_SWAP: {
        BTN_SINGLE: BTN_TRIG_DOUBLE,
        BTN_DOUBLE: BTN_TRIG_SINGLE,
        BTN_TRIPLE: BTN_TRIG_TRIPLE,
        BTN_QUAD: BTN_TRIG_QUAD,
        BTN_PENTA: BTN_TRIG_PENTA,
        BTN_HOLD: BTN_TRIG_HOLD,
    },
    BUTTONMODE_SINGLE: {
        BTN_SINGLE: BTN_TRIG_SINGLE,
        BTN_DOUBLE: BTN_TRIG_NONE,
        BTN_TRIPLE: BTN_TRIG_NONE,
        BTN_QUAD: BTN_TRIG_NONE,
        BTN_PENTA: BTN_TRIG_NONE,
        BTN_HOLD: BTN_TRIG_NONE,
    },
}


@attr.s(slots=True, frozen=True)
class TasmotaButtonTriggerConfig(TasmotaTriggerConfig):
    """Tasmota switch configuation."""

    @classmethod
    def from_discovery_message(
        cls, config: dict, idx: int
    ) -> list[TasmotaButtonTriggerConfig]:
        """Instantiate from discovery message."""
        mqtt_buttons = config[CONF_OPTIONS][OPTION_MQTT_BUTTONS]
        single_buttons = config[CONF_OPTIONS][OPTION_BUTTON_SINGLE]
        swap_buttons = config[CONF_OPTIONS][OPTION_BUTTON_SWAP]
        buttonmode = BUTTONMODE_NONE
        if mqtt_buttons and config[CONF_BUTTON][idx]:
            if single_buttons:
                buttonmode = BUTTONMODE_SINGLE
            elif swap_buttons:
                buttonmode = BUTTONMODE_SWAP
            else:
                buttonmode = BUTTONMODE_NORMAL

        triggers = BUTTONMODE_MAP[buttonmode]
        configs = []
        for event, trigger_type in triggers.items():
            configs.append(
                cls(
                    mac=config[CONF_MAC],
                    event=event,
                    idx=idx,
                    source="button",
                    subtype=f"button_{idx+1}",
                    trigger_topic=get_topic_stat_result(config),
                    type=trigger_type,
                )
            )
        return configs

    @property
    def is_active(self) -> bool:
        """Return if the trigger is active."""
        return self.type != BTN_TRIG_NONE

    @property
    def trigger_id(self) -> str:
        """Return trigger id."""
        return f"{self.mac}_button_{self.idx+1}_{self.event}"


class TasmotaButtonTrigger(TasmotaTrigger):
    """Representation of a Tasmota button trigger."""

    cfg: TasmotaButtonTriggerConfig

    def _trig_message_received(self, msg: ReceiveMessage) -> None:
        """Handle new MQTT messages."""
        event = get_value_by_path(msg.payload, [f"Button{self.cfg.idx+1}", RSLT_ACTION])
        if event == self.cfg.event and self._on_trigger_callback:
            self._on_trigger_callback()



"""Tasmota config validation."""
from __future__ import annotations

from typing import Any, List, TypeVar, Union

import voluptuous as vol

# typing typevar
T = TypeVar("T")  # pylint: disable=invalid-name

bit = vol.All(vol.Coerce(int), vol.Range(min=0, max=1))


positive_int = vol.All(  # pylint: disable=invalid-name
    vol.Coerce(int), vol.Range(min=0)
)


def ensure_list(value: Union[T, List[T], None]) -> List[T]:
    """Wrap value in list if it is not one."""
    if value is None:
        return []
    return value if isinstance(value, list) else [value]


def optional_string(value: Any) -> str | None:
    """Coerce value to string, except for None."""
    if value is None:
        return None
    if isinstance(value, (list, dict)):
        raise vol.Invalid("value should be a string")

    return str(value)


def string(value: Any) -> str:
    """Coerce value to string, except for None."""
    if value is None:
        raise vol.Invalid("string value is None")
    if isinstance(value, (list, dict)):
        raise vol.Invalid("value should be a string")

    return str(value)



"""Tasmota binary sensor."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable

import attr

from .const import AUTOMATION_TYPE_TRIGGER
from .mqtt import ReceiveMessage, TasmotaMQTTClient

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True, frozen=True)
class TasmotaTriggerConfig(ABC):
    """Tasmota trigger configuation."""

    event: str = attr.ib()
    idx: int = attr.ib()
    mac: str = attr.ib()
    subtype: str = attr.ib()
    source: str = attr.ib()
    trigger_topic: str = attr.ib()
    type: str = attr.ib()

    @property
    @abstractmethod
    def is_active(self) -> int:
        """Return if the trigger is active."""

    @property
    @abstractmethod
    def trigger_id(self) -> str:
        """Return trigger id."""


class TasmotaTrigger:
    """Representation of a Tasmota trigger."""

    def __init__(
        self, config: TasmotaTriggerConfig, mqtt_client: TasmotaMQTTClient, **_kwds: Any
    ):
        """Initialize."""
        self._sub_state: dict | None = None
        self.cfg = config
        self._mqtt_client = mqtt_client
        self._on_trigger_callback: Callable | None = None

    def config_same(self, new_config: TasmotaTriggerConfig) -> bool:
        """Return if updated config is same as current config."""
        return self.cfg == new_config

    def config_update(self, new_config: TasmotaTriggerConfig) -> None:
        """Update config."""
        self.cfg = new_config

    def set_on_trigger_callback(self, on_trigger_callback: Callable) -> None:
        """Set callback for triggere."""
        self._on_trigger_callback = on_trigger_callback

    def _trig_message_received(self, msg: ReceiveMessage) -> None:
        """Handle new MQTT messages."""

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

        topics = {
            "trigger_topic": {
                "event_loop_safe": True,
                "topic": self.cfg.trigger_topic,
                "msg_callback": lambda msg: self._trig_message_received(  # pylint: disable=unnecessary-lambda
                    msg
                ),
            }
        }

        self._sub_state = await self._mqtt_client.subscribe(
            self._sub_state,
            topics,
        )

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe to all MQTT topics."""
        self._sub_state = await self._mqtt_client.unsubscribe(self._sub_state)

    @property
    def automation_type(self) -> str:
        """Return the automation type."""
        return AUTOMATION_TYPE_TRIGGER



"""Tasmota status sensor."""
from __future__ import annotations

import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from typing import Any

import attr

from .const import (
    CONF_DEVICENAME,
    CONF_IP,
    CONF_MAC,
    PERCENTAGE,
    SENSOR_STATUS_IP,
    SENSOR_STATUS_LAST_RESTART_TIME,
    SENSOR_STATUS_LINK_COUNT,
    SENSOR_STATUS_MQTT_COUNT,
    SENSOR_STATUS_RESTART_REASON,
    SENSOR_STATUS_RSSI,
    SENSOR_STATUS_SIGNAL,
    SENSOR_STATUS_SSID,
    SENSOR_STATUS_VERSION,
    SIGNAL_STRENGTH_DECIBELS_MILLIWATT,
)
from .entity import TasmotaAvailability, TasmotaEntity
from .mqtt import ReceiveMessage
from .sensor import TasmotaBaseSensorConfig
from .utils import (
    config_get_state_offline,
    config_get_state_online,
    get_topic_command_status,
    get_topic_stat_status,
    get_topic_tele_state,
    get_topic_tele_will,
    get_value_by_path,
)

_LOGGER = logging.getLogger(__name__)

# 14:45:37 MQT: tasmota_B94927/tele/HASS_STATE = {
#  "Version":"9.0.0.1(tasmota)",               stat/STATUS2:"StatusFWR"."Version"
#  "BuildDateTime":"2020-10-08T21:38:21",      stat/STATUS2:"StatusFWR"."BuildDateTime"
#  "Module or Template":"Generic",             <NONE>
#  "RestartReason":"Software/System restart",  stat/STATUS1:"StatusPRM"."RestartReason"
#  "Uptime":"1T17:04:28",                      stat/STATUS11:"StatusSTS"."Uptime"; tele/STATE:"Uptime"
#  "Hostname":"tasmota_B94927",                stat/STATUS5:"StatusNET":"Hostname"
#  "IPAddress":"192.168.0.114",                stat/STATUS5:"StatusNET":"IPAddress"
#  "RSSI":"100",                               stat/STATUS11:"StatusSTS":"RSSI"; tele/STATE:"RSSI"
#  "Signal (dBm)":"-49",                       stat/STATUS11:"StatusSTS":"Signal"; tele/STATE:"Signal"
#  "WiFi LinkCount":1,                         stat/STATUS11:"StatusSTS":"LinkCount"; tele/STATE:"LinkCount"
#  "WiFi Downtime":"0T00:00:03",               stat/STATUS11:"StatusSTS":"Downtime"; tele/STATE:"Downtime"
#  "MqttCount":1,                              stat/STATUS11:"StatusSTS":"MqttCount"; tele/STATE:"MqttCount"
#  "LoadAvg":19                                stat/STATUS11:"StatusSTS":"LoadAvg"; tele/STATE:"LoadAvg"
# }

SENSORS = [
    SENSOR_STATUS_IP,
    SENSOR_STATUS_LAST_RESTART_TIME,
    SENSOR_STATUS_LINK_COUNT,
    SENSOR_STATUS_MQTT_COUNT,
    SENSOR_STATUS_RESTART_REASON,
    SENSOR_STATUS_RSSI,
    SENSOR_STATUS_SIGNAL,
    SENSOR_STATUS_SSID,
    SENSOR_STATUS_VERSION,
]

NAMES = {
    SENSOR_STATUS_IP: "IP",
    SENSOR_STATUS_LAST_RESTART_TIME: "Last Restart Time",
    SENSOR_STATUS_LINK_COUNT: "WiFi Connect Count",
    SENSOR_STATUS_MQTT_COUNT: "MQTT Connect Count",
    SENSOR_STATUS_RESTART_REASON: "Restart Reason",
    SENSOR_STATUS_RSSI: "RSSI",
    SENSOR_STATUS_SIGNAL: "Signal",
    SENSOR_STATUS_SSID: "SSID",
    SENSOR_STATUS_VERSION: "Firmware Version",
}

SINGLE_SHOT = [
    SENSOR_STATUS_LAST_RESTART_TIME,
    SENSOR_STATUS_RESTART_REASON,
    SENSOR_STATUS_VERSION,
]

STATE_PATHS: dict[str, list[str | int]] = {
    SENSOR_STATUS_LINK_COUNT: ["Wifi", "LinkCount"],
    SENSOR_STATUS_MQTT_COUNT: ["MqttCount"],
    SENSOR_STATUS_RSSI: ["Wifi", "RSSI"],
    SENSOR_STATUS_SIGNAL: ["Wifi", "Signal"],
}

STATUS_PATHS: dict[str, list[str | int]] = {
    SENSOR_STATUS_LAST_RESTART_TIME: ["StatusSTS", "UptimeSec"],
    SENSOR_STATUS_LINK_COUNT: ["StatusSTS", "Wifi", "LinkCount"],
    SENSOR_STATUS_MQTT_COUNT: ["StatusSTS", "MqttCount"],
    SENSOR_STATUS_RESTART_REASON: ["StatusPRM", "RestartReason"],
    SENSOR_STATUS_RSSI: ["StatusSTS", "Wifi", "RSSI"],
    SENSOR_STATUS_SIGNAL: ["StatusSTS", "Wifi", "Signal"],
    SENSOR_STATUS_SSID: ["StatusSTS", "Wifi", "SSId"],
    SENSOR_STATUS_VERSION: ["StatusFWR", "Version"],
}

STATUS_TOPICS = {
    SENSOR_STATUS_LAST_RESTART_TIME: 11,
    SENSOR_STATUS_LINK_COUNT: 11,
    SENSOR_STATUS_MQTT_COUNT: 11,
    SENSOR_STATUS_RESTART_REASON: 1,
    SENSOR_STATUS_RSSI: 11,
    SENSOR_STATUS_SIGNAL: 11,
    SENSOR_STATUS_SSID: 11,
    SENSOR_STATUS_VERSION: 2,
}

QUANTITY = {
    SENSOR_STATUS_IP: SENSOR_STATUS_IP,
    SENSOR_STATUS_LAST_RESTART_TIME: SENSOR_STATUS_LAST_RESTART_TIME,
    SENSOR_STATUS_LINK_COUNT: SENSOR_STATUS_LINK_COUNT,
    SENSOR_STATUS_MQTT_COUNT: SENSOR_STATUS_MQTT_COUNT,
    SENSOR_STATUS_RESTART_REASON: SENSOR_STATUS_RESTART_REASON,
    SENSOR_STATUS_RSSI: SENSOR_STATUS_RSSI,
    SENSOR_STATUS_SIGNAL: SENSOR_STATUS_SIGNAL,
    SENSOR_STATUS_SSID: SENSOR_STATUS_SSID,
    SENSOR_STATUS_VERSION: SENSOR_STATUS_VERSION,
}

UNITS = {
    SENSOR_STATUS_IP: None,
    SENSOR_STATUS_LAST_RESTART_TIME: None,
    SENSOR_STATUS_LINK_COUNT: None,
    SENSOR_STATUS_MQTT_COUNT: None,
    SENSOR_STATUS_RESTART_REASON: None,
    SENSOR_STATUS_RSSI: PERCENTAGE,
    SENSOR_STATUS_SIGNAL: SIGNAL_STRENGTH_DECIBELS_MILLIWATT,
    SENSOR_STATUS_SSID: None,
    SENSOR_STATUS_VERSION: None,
}


@attr.s(slots=True, frozen=True)
class TasmotaStatusSensorConfig(TasmotaBaseSensorConfig):
    """Tasmota Status Sensor configuration."""

    poll_topic: str = attr.ib()
    sensor: str = attr.ib()
    state: str = attr.ib()
    state_topic: str = attr.ib()
    status_topic: str = attr.ib()

    @classmethod
    def from_discovery_message(
        cls, config: dict, platform: str
    ) -> list[TasmotaStatusSensorConfig]:
        """Instantiate from discovery message."""
        sensors = [
            cls(
                endpoint="status_sensor",
                idx=None,
                friendly_name=f"{config[CONF_DEVICENAME]} {NAMES[sensor]}",
                mac=config[CONF_MAC],
                platform=platform,
                poll_payload=str(STATUS_TOPICS.get(sensor)),
                poll_topic=get_topic_command_status(config),
                availability_topic=get_topic_tele_will(config),
                availability_offline=config_get_state_offline(config),
                availability_online=config_get_state_online(config),
                sensor=sensor,
                state=config[CONF_IP] if sensor == SENSOR_STATUS_IP else None,
                state_topic=get_topic_tele_state(config),
                status_topic=get_topic_stat_status(config, STATUS_TOPICS.get(sensor)),
            )
            for sensor in SENSORS
        ]
        return sensors

    @property
    def unique_id(self) -> str:
        """Return unique_id."""
        return f"{self.mac}_{self.platform}_{self.endpoint}_{self.sensor}"


class TasmotaStatusSensor(TasmotaAvailability, TasmotaEntity):
    """Tasmota Status sensors."""

    _cfg: TasmotaStatusSensorConfig

    def __init__(self, **kwds: Any):
        """Initialize."""
        self._sub_state: dict | None = None
        self._sub_state_lock = asyncio.Lock()
        super().__init__(**kwds)

    async def _poll_status(self) -> None:
        """Poll for status."""
        await self.subscribe_topics()
        await self._mqtt_client.publish_debounced(
            self._cfg.poll_topic, self._cfg.poll_payload
        )

    async def poll_status(self) -> None:
        """Poll for status."""
        await self._poll_status()

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

        def state_message_received(msg: ReceiveMessage) -> None:
            """Handle new MQTT state messages."""
            if not self._on_state_callback:
                return

            try:
                payload = json.loads(msg.payload)
            except (json.decoder.JSONDecodeError):
                return

            state = None
            if msg.topic == self._cfg.state_topic:
                state = get_value_by_path(payload, STATE_PATHS[self._cfg.sensor])
            else:
                state = get_value_by_path(payload, STATUS_PATHS[self._cfg.sensor])
            if state is not None:
                if self._cfg.sensor in SINGLE_SHOT:
                    asyncio.create_task(self._unsubscribe_state_topics())
                if self._cfg.sensor == SENSOR_STATUS_LAST_RESTART_TIME:
                    state = datetime.now(timezone.utc) - timedelta(seconds=int(state))
                self._on_state_callback(state)

        availability_topics = self.get_availability_topics()
        topics = {}
        if self._cfg.sensor in STATE_PATHS:
            # Periodic state update (tele/STATE)
            topics["state_topic"] = {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic,
                "msg_callback": state_message_received,
            }
        if self._cfg.sensor in STATUS_PATHS:
            # Polled state update (stat/STATUS#)
            topics["status_topic"] = {
                "event_loop_safe": True,
                "topic": self._cfg.status_topic,
                "msg_callback": state_message_received,
            }
        topics = {**topics, **availability_topics}

        async with self._sub_state_lock:
            self._sub_state = await self._mqtt_client.subscribe(
                self._sub_state,
                topics,
            )
        if self._cfg.state and self._on_state_callback:
            self._on_state_callback(self._cfg.state)

    async def _unsubscribe_state_topics(self) -> None:
        """Unsubscribe from state topics."""
        availability_topics = self.get_availability_topics()
        async with self._sub_state_lock:
            self._sub_state = await self._mqtt_client.subscribe(
                self._sub_state,
                availability_topics,
            )

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe from all MQTT topics."""
        async with self._sub_state_lock:
            self._sub_state = await self._mqtt_client.unsubscribe(self._sub_state)

    @property
    def quantity(self) -> str:
        """Return the sensor's quantity (speed, mass, etc.)."""
        return QUANTITY[self._cfg.sensor]

    @property
    def unit(self) -> str | None:
        """Return the unit this state is expressed in."""
        return UNITS[self._cfg.sensor]
